#!/usr/bin/env python
# coding: utf-8

"""Various small mathematical functions

"""

import logging

import numpy
import numpy.ma
import numpy.linalg
import scipy
import scipy.optimize
import scipy.stats
import scipy.odr

import sklearn.linear_model
import pint

from . import tools
from .meta import expanddoc
from .units import ureg

inputs = """:param z: Height
    :type z: ndarray
    :param q: Quantity, dim 0 must be height.
    :type q: ndarray
    :param ignore_negative: Set negative values to 0
    :type ignore_negative: bool"""

# Make my own exception because scipy ODR doesn't raise, just gives a
# message in a list…
class ODRFitError(ArithmeticError):
    pass

@expanddoc
def layer2level(z, q, ignore_negative=False):
    """Converts layer to level. First dim. must be height.

    {inputs}
    :returns: Level-valued quantity.
    """
    dz = z[1:, ...] - z[:-1, ...]
    if ignore_negative:
        q[q<0]=0
    y_avg = (q[1:, ...] + q[:-1, ...])/2
    return (y_avg * numpy.atleast_2d(dz).T)

@expanddoc
def integrate_with_height(z, q, ignore_negative=False):
    """Calculate vertically integrated value

    {inputs}
    :returns: Vertically integrated value
    """

    return layer2level(z, q, ignore_negative).sum(0)

@expanddoc
def cum_integrate_with_height(z, q, ignore_negative=False):
    """Calculate cumulative integral with height

    {inputs}
    :returns: Cumulatively vertically integrated value
    """
    return layer2level(z, q, ignore_negative).cumsum(0)

#@tools.validator # comment out because fails for *args
def average_position_sphere(*args: (lambda a: len(a) in (1,2))):
    """Calculate the average position for a set of angles on a sphere

    This is quite imprecise, errors can be dozens of km.  For more
    advanced calculations, use proj4/pyproj.

    Input can be either:

    :param lat: Vector of latitudes
    :param lon: Vector of longitudes

    Or:

    :param locs: Nx2 ndarray with lats in column 0, lons in column 1
    """

    if len(args) == 1:
        locs = args[0]
        lat = locs[0, :]
        lons = locs[1, :]
    elif len(args) == 2:
        lat, lon = args

    X = numpy.cos(lat) * numpy.cos(lon)
    Y = numpy.cos(lat) * numpy.sin(lon)
    Z = numpy.sin(lat)

    xm = X.mean()
    ym = Y.mean()
    zm = Z.mean()

    lonm = numpy.arctan2(ym, xm)
    latm = numpy.arctan2(zm, numpy.sqrt(xm**2+ym**2))

    return (latm, lonm)

def linear_interpolation_matrix(x_old, x_new):
    """Get transformation matrix for linear interpolation.

    This is denoted by W in Calisesi, Soebijanta and Van Oss (2005).

    Does note extrapolate; values outside the range are equal to the
    outermost values.

    :param x_old: Original 1-D grid
    :param x_new: New 1-D grid for interpolation
    :returns ndarray W: Interpolation transformation matrix.
    """
    
    W = numpy.vstack(
        [numpy.interp(x_new, x_old, numpy.eye(x_old.size)[i, :],
            left=numpy.nan, right=numpy.nan)    
            for i in range(x_old.size)])
    #return W
    return W.T


#    return numpy.vstack(
#        [scipy.interpolate.InterpolatedUnivariateSpline(
#            x_old, eye(x_old.size)[i, :])(x_new) 
#                for i in range(x_old.size)])

def regrid_ak(A, z_old, z_new, cut=False):
    """Regrid averaging kernel matrix.

    Actual regridding done in apply_W_A, following Calisesi, Soebijanta
    and Van Oss (2005).

    :param A: Original averaging kernel
    :param z_old: Original z-grid
    :param z_new: New z-grid
    :param bool cut: Cut off, i.e. flag, when any z in the new grid is
        outside the old grid.
    :returns: (New averaging kernel, W)
    """

    if cut:
        # make sure we take care of flagged data
        valid = ~(A<-10).all(0)

        # Not on the new one!  We must output the same size for A every
        # time.
#        new_outside = ((z_new > numpy.nanmax(z_old)) | 
#                       (z_new < numpy.nanmin(z_old)))
        #               ~numpy.isfinite(z_old))
        z_old_valid = numpy.isfinite(z_old)
#        if z_old[z_old_valid].max() < z_new.max():
#            raise ValueError("z_new not a subset of z_old!")
        #W = linear_interpolation_matrix(z_old[z_old_valid], z_new[~new_outside])
        #W = linear_interpolation_matrix(z_old[z_old_valid], z_new)
        # Keep full W (unflagged) because I want to put them all in a
        # single ndarray later
        # TODO: retry with masked arrays after the bugfixes
        W = linear_interpolation_matrix(z_old, z_new)
        #z_new_ok = (z_new > z_old.min()) & (z_new < z_old.max())
        z_new_ok = numpy.isfinite(W).all(1)
        A_new = numpy.zeros(shape=(z_new.shape[0], z_new.shape[0]))
        A_new.fill(numpy.nan)
        # The first one seems to not work... the second one does
        #A_new[z_new_ok, :][:, z_new_ok] = apply_W_A(
        A_new[numpy.ix_(z_new_ok, z_new_ok)] = apply_W_A(
                W[:, z_old_valid][z_new_ok, :],
                A[z_old_valid, :][:, z_old_valid])
#        A_new[outside, outside] = numpy.nan
        return (A_new, W)
    else:
        W = linear_interpolation_matrix(z_old, z_new)
        return (apply_W_A(W, A), W)

def regrid_matrix(A, z_old, z_new):
    """Regrid single matrix between grids.

    Do not use for averaging kernels!

    :param A:
    :param z_old:
    :param nd-array z_new: 1-D array
    """

    if z_old.shape[1] != A.shape[1]:
        raise ValueError("Shapes dont match")
    scipy.interpolate.interp1d
    A_new = numpy.zeros(shape=(z_old.shape[0], z_new.shape[0]))
    for i in range(z_old.shape[0]):
        for x in range(A.shape[0]):
            ip = scipy.interpolate.interp1d(z_old[i, :], A[x, :],
                bounds_error=False)
            A_new[i, :] = ip(z_new)
    return A_new

def apply_W_A(W, A):
    """Regrid averaging kernel matrix using W

    If interpolation matrix W is already calculated, apply to averaging
    kernel matrix here.

    This follows the methodology outlined by Calisesi, Soebijanta and Van
    Oss (2005).
    """

    Wstar = numpy.linalg.pinv(W)
    return W.dot(A).dot(Wstar)

def convert_ak_ap2vmr(AKx, aprf):
    """Convert averaging kernel from SFIT4 units to vmr

    :param AKx: Averaging kernel from SFIT4
    :param aprf: A-priori
    :returns: Averaging kernel in VMR units
    """

    # Source: e-mail Stephanie 2014-06-17

    return numpy.diag(1/aprf).dot(AKx).dot(numpy.diag(aprf))

def smooth_profile(xh, ak, xa):
    """Calculated smoothed profile.

    Calculate a smoothed profile following Rodgers and Connor (2003).

    :param xh: High-resolution profile
    :param ak: Low-resolution averaging kernel [VMR]
    :param xa: Low-resolution a priori profile
    """

    OK = (numpy.isfinite(xa) &
          numpy.isfinite(numpy.diag(ak)) &
          numpy.isfinite(xh))
    xs = numpy.zeros_like(xa)
    xs.fill(numpy.nan)
    xs[OK] = xa[OK] + ak[numpy.ix_(OK,OK)].dot(xh[OK] - xa[OK])
    return xs

def mad(x, masked=False):
    """Median absolute deviation
    """

    med = numpy.ma.median if masked else numpy.median

    return med(numpy.abs(x - med(x)))

def get_transformation_matrix(f, n):
    """Obtain standard matrix for the linear transformaton

    For a given linear function taking a vector valued argument, obtain
    the standard matrix for the linear transformation.

    See Lay (2003), Linear Algebra and its Transformations, 3rd edition,
    Theorem 10 (page 83).

    :param callable f: Function for which to get the transformation
        matrix.  This might be a function or a functools.partial object,
        for example.  This function should take as input a vector of
        length n and return as output a vector of length m (m>0).
        Of course, this function should be linear.
    :param int n: Size of transformation matrix needed
    :returns: (m x n) transformation matrix.
    """

    I = numpy.eye(n)
    return numpy.hstack([f(I[:, i:(i+1)]) for i in range(n)])

def calc_y_for_srf_shift(Δλ, y_master, srf0, L_spectral_db, f_spectra, y_ref,
                           unit=ureg.um,
                           regression_type=sklearn.linear_model.LinearRegression,
                           regression_args={"fit_intercept": True},
                           predict_quantity="bt",
                           u_y_ref=None,
                           u_y_target=None):
    """Calculate radiances or BTs estimating y_target from y_master assuming srf0 shifts by Δλ

    Try to estimate y_target from y_master, assuming that y_master are
    bts or radiances corresponding to spectral response function srf0.
    For the estimate, use a database described by L_spectral_db and
    f_spectra.  This database will be used to train a regression from
    the original bt (y_master, due to srf0) to a shifted bt (due to
    srf0 shifted by Δλ).  The regression is trained using the spectral
    database (L_spectral_db, f_spectra, bt_ref).  This function
    then applies the regression to y_master.

    This function is designed to be called by
    `:func:calc_cost_for_srf_shift`, which in turn is designed
    to be called repeatedly within an optimisation framework (see
    `:func:estimate_srf_shift`).  Therefore, as much as possible is
    precalculated before calling this.  Hence, you also need to pass
    bt_ref, which equals integrated brightness temperatures for
    the reference satellite, corresponding to f_spectra and L_spectral_db.

    This estimate considers one channel at a time.  It may be more optimal
    to estimate multiple channels at a time.  This is to be done.

    Arguments:
        
        Δλ (Quantity or float): shift in SRF.  Will be converted to the
            unit (typically µm or nm) from the pint user registry (see
            later argument).  Scalar.
        y_master (Quantity ndarray, N×k): Brightness temperatures [K] or
            radiances [radiance units] for reference satellite.  N
            samples, k channels.  Quantity must be consistent with the one
            described by predict_quantity.  These are used in the actual
            prediction; the training is performed with y_ref.  For
            example, if we are predicting NOAA18-8 from NOAA19-1-12 in a
            test setting, y_master would correspond to NOAA19-1-12 test
            data, and y_ref would be the same for training data.  In the
            real world, y_ref is still simulated, but y_master are actual
            measurements.
        srf0 (`:func:pyatmlab.physics.SRF`): SRF relative to which
            the shift is to be calculated.
        L_spectral_db (ndarray M×l): Database of spectra (such as from IASI)
            to use.  Should be in spectral radiance per frequency units [W
            / (m^2 sr Hz)].  M spectra with l radiances each.
        f_spectra (Quantity ndarray l): frequencies corresponding to
            L_spectral_db [Hz].  1-D with length l.
        y_ref (Quantity ndarray M×k): Predictands used to train regressions,
            i.e. the training database.  This information follows directly from
            L_spectral_db and SRFs on the reference satellite, but it is
            an expensive calculation so should be pre-calculated.  If
            predicting from the same satellite, at least one channel will
            correspond to srf0, such that

                # in case of radiances
                L = srf0.integrate_radiances(f_spectra, L_spectral_db)
                # in case of BTs
                bt = srf0.channel_radiance2bt(L)

            but this is not the case if one satellite is predicted from
            another.
        unit (Unit): unit from pint unit registry.  Defaults to ureg.um.
        regression_type (scikit-learn regressor): Type of regression.
            Defaults to sklearn.linear_model.LinearRegression.  Other good
            option would be sklearn.cross_decomposition.PLSRegression.
            As long as regression_type(**regression_args) behaves like
            those two (with .fit and .predict), it should be OK.
        regression_args (dict): Keyword arguments to pass on to regressor.
            For example, for sklearn.linear_model.LinearRegression you
            would want to at least pass `{"fit_intercept": True}`.  For
            sklearn.cross_decomposition.PLSRegression you might use
            `{"n_components": 9, "scale": False}`.  Please refer to
            scikit-learn documentation.


    Returns:

        ndarray with estimates for shifted y_master or y_master values
    """
    try:
        Δλ = Δλ.to(unit)
    except AttributeError:
        Δλ = ureg.Quantity(Δλ, unit)
    srf_sh = srf0.shift(Δλ)
    L_target = srf_sh.integrate_radiances(f_spectra, L_spectral_db)
    if predict_quantity == "bt":
        y_target = srf_sh.channel_radiance2bt(L_target)
    elif predict_quantity == "radiance":
        y_target = L_target
    else:
        raise ValueError("Invalid predict_quantity: {:s}".format(predict_quantity))

    # sklearn wants 2-D inputs even when there is only a single predictand
    # atleast_2d makes it (1, N), I need (N, 1), but transposing the
    # result of atleast_2d would be wrong if the array was already 2D to
    # begin with
    if y_master.ndim == 1:
        y_master = y_master[:, numpy.newaxis]

    if y_ref.ndim == 1:
        y_ref = y_ref[:, numpy.newaxis]

    #clf = sklearn.linear_model.LinearRegression(fit_intercept=True)
    if issubclass(regression_type, sklearn.base.RegressorMixin):
        clf = regression_type(**regression_args)
        clf.fit(y_ref.m, y_target.m)
        return ureg.Quantity(clf.predict(y_master.m).squeeze(), y_master.u)
    elif issubclass(regression_type, scipy.odr.odrpack.ODR):
        try:
            sx = u_y_ref.to(y_ref.u, "radiance").m.squeeze()
            sy = u_y_target.to(u_y_target.u, "radiance").m.squeeze()
        except pint.DimensionalityError: # probably noise in other unit
            # here, those are really only used as weights, so same-unit
            # is not essential?  Otherwise I need to convert them, a
            # non-trivial step.
            sx = u_y_ref.m.squeeze()
            sy = u_y_target.m.squeeze()

        # meaningful initial guess: new value equal to old channel, rest 0
        # β0 has a offset and then one slope for each channel
        # don't set to zero, says odrpack guide section 1.E.
        β0 = numpy.zeros(shape=(y_ref.shape[1]+1,))+0.001
        samech = abs((y_target[:, numpy.newaxis] - y_ref)).mean(0).argmin()
        β0[samech+1] = 1
        if (sx==0).any() or (sy==0).any():
            mydata = scipy.odr.RealData(y_ref.m.T, y_target.m)
        else:
            mydata = scipy.odr.RealData(y_ref.m.T, y_target.m, sx=sx, sy=sy)
        myodr = scipy.odr.ODR(mydata, scipy.odr.multilinear, beta0=β0)
        myout = myodr.run()
        if not any(x in myout.stopreason for x in
            {"Sum of squares convergence",
             "Iteration limit reached",
             "Parameter convergence",
             "Both sum of squares and parameter convergence"}):
            raise ODRFitError("ODR fitting did not converge.  "
                "Stop reason: {:s}".format(myout.stopreason[0]))
        return ureg.Quantity(myout.beta[0], y_master.u) + (
            myout.beta[numpy.newaxis, 1:] * y_master).sum(1)
    else:
        raise ValueError("Unknown type of regression!")
#    (slope, intercept, r_value, p_value, stderr) = scipy.stats.linregress(
#            bt_ref, bt_target)
#    
#    return intercept*bt_ref.u + slope*bt_master

def calc_cost_for_srf_shift(Δλ, y_master, y_target, srf0,
                            L_spectral_db, f_spectra, y_ref,
                            unit=ureg.um,
                            regression_type=sklearn.linear_model.LinearRegression,
                            regression_args={"fit_intercept": True},
                            cost_mode="total",
                            predict_quantity="bt",
                            u_y_ref=None,
                            u_y_target=None):
    """Calculate cost function estimating y_target from y_master assuming
    srf0 shifts by Δλ

    Try to estimate how well we can estimate y_target from y_master,
    assuming that y_master are radiances or BTs
    corresponding to spectral response function srf0.  For the estimate,
    use a database described by L_spectral_db and f_spectra.

    This function is designed to be called repeatedly within an
    optimisation framework (see `:func:estimate_srf_shift`).  Therefore,
    as much as possible is precalculated before calling this.  Hence, you
    also need to pass y_ref, which equals integrated radiances for the
    reference satellite, corresponding to f_spectra and L_spectral_db,
    even though y_ref is redundant information.

    This estimate considers one channel at a time.  It may be more optimal
    to estimate multiple channels at a time.  This is to be done.

    For example, y_master could be radiances/BTs corresponding to channel 1
    on MetOp-A, srf0 the SRF for channel 1 on MetOp-A.  y_target
    could be radiances or BTs for NOAA-19, suspected to be approximated by
    shifting srf0 by an unknown amount.  Δλ would be a shift amount
    to attempt, such as +30 nm.  We could then continue trying different
    values for Δλ until we minimise the cost.

    The two alternative cost functions are:

        C₁ = \sum_{i=1}^N (y_est,i - y_ref,i)^2

    and

        C₂ = \sum_{i=1}^N (y_est,i - y_ref,i - <y_est,i - y_ref,i>)^2

    Arguments:
        
        Δλ (Quantity): shift in SRF.
        y_master (ndarray): Radiances [radiance units] or BTs [K] for
            reference satellite.  This comes from either the testing data,
            or actual measurements.  Quantity must be consistent with what is
            contained in predict_quantity.
        y_target (ndarray): Radiances or BTs for other satellite.  This
            comes from either the testing data (calculated by shifting
            srf0 by an amount you haven't told me, but I mean to recover),
            or from actual measurements.
        srf0 (`:func:pyatmlab.physics.SRF`): SRF corresponding to
            zero shift, relative to which the shift is estimated.
        L_spectral_db (ndarray N×p): Database of spectra (such as from IASI)
            to use.  Should be in SI spectral radiance per frequency units [W
            / (m^2 sr Hz)], regardless of predict_quantity.
        f_spectra (ndarray N): frequencies corresponding to L_spectral_db [Hz]
        y_ref: Radiances or brightness temperatures corresponding to 
            L_spectral_db and f_spectra [K], for all channels to be used.
            This is essential the training database.  It comes from
            simulations regardless as to we are applying this as a test or
            in the real world.
        unit (Unit): unit from pint unit registry.  Defaults to ureg.um.
        regression_type (scikit-learn regressor): Type of regression.
            Defaults to sklearn.linear_model.LinearRegression.  Other good
            option would be sklearn.cross_decomposition.PLSRegression.
            As long as regression_type(**regression_args) behaves like
            those two (with .fit and .predict), it should be OK.  Some
            other cases are also understood, most notably scipy.odr.ODR.
        regression_args (dict): Keyword arguments to pass on to regressor.
            For example, for sklearn.linear_model.LinearRegression you
            would want to at least pass `{"fit_intercept": True}`.  For
            sklearn.cross_decomposition.PLSRegression you might use
            `{"n_components": 9, "scale": False}`.  Please refer to
            scikit-learn documentation for details.
        cost_mode (str): How to estimate the cost.  Can be "total"
            (default), which calculates C₁, or "anomalies", which
            calculates C₂, according to their definitions above.
        predict_quantity (str): "bt" (default) or "radiance"
        u_y_ref (ndarray): Uncertainties on y_ref.  Used by some
            regression types.
        u_y_target (ndarray): Uncertainties on y_target.

    Returns:
        
        cost (float): Value of estimated cost function.  Unitless.
    """
    y_estimate = calc_y_for_srf_shift(Δλ, y_master, srf0,
            L_spectral_db, f_spectra, y_ref, unit,
            regression_type=regression_type,
            regression_args=regression_args,
            predict_quantity=predict_quantity,
            u_y_ref=u_y_ref,
            u_y_target=u_y_target)
    diffs = y_target - y_estimate
    if cost_mode.lower() == "anomalies":
        diffs -= (y_target - y_estimate).mean()
    elif cost_mode.lower() != "total":
        raise ValueError("Invalid cost mode: {:s}.  I understand 'total' "
                         "and 'anomalies'".format(cost_mode))
    cost = (diffs**2).sum() / (diffs.size * y_target.mean()**2)
    return cost

def estimate_srf_shift(y_master, y_target, srf0, L_spectral_db, f_spectra,
        y_ref,
        regression_type, regression_args,
        optimiser_func, optimiser_args,
        cost_mode,
        predict_quantity="bt",
        u_y_ref=None,
        u_y_target=None,
        **solver_args):
    """Estimate shift in SRF from pairs of brightness temperatures

    From pairs of brightness temperatures, estimate what SRF shifts
    minimises observed BT differences.

    Arguments:
        
        y_master (ndarray): Radiances or BTs for reference satellite.
            Unit must be consistent with what you tell me in
            predict_quantity.
        y_target (ndarray): Radiances or BTs for other satellite
        srf0 (`:func:pyatmlab.physics.SRF`): SRF for reference satellite
        L_spectral_db (ndarray N×p): Database of spectra (such as from IASI)
            to use.  Should ALWAYS be in spectral radiance per frequency
            units, regardless of what predict_quantity is.
        f_spectra (ndarray N): spectrum describing frequencies
            corresponding to `L_spectral_db`.  In Hz.
        y_ref: Reference BT or radiance
        regression_type (scikit-learn regressor): As for
            `:func:calc_cost_for_srf_shift`.
        regression_args (scikit-learn regressor): As for
            `:func:calc_cost_for_srf_shift`.
        optimiser_func (function): Function implementing optimising.
            Should take as a first argument a function to be optimised,
            remaining arguments taken from optimiser_args.
            Should return an instance of
            `:func:scipy.optimize.optimize.OptimizeResult`.  You probably
            want to select a function from `:module:scipy.optimize`, such
            as `:func:scipy.optimize.basinhopping` when there may be
            multiple minima, or `:func:scipy.optimize,minimize_scalar`
            when you expect only one.
        optimiser_args (dict): Keyword arguments to pass on to
            `optimiser_func`.
        cost_mode (str): As for `:func_calc_cost_for_srf_shift`.
        predict_quantity (str): Whether to perform prediction in
            "radiance" or in "bt".  y_master and y_target should be in
            units corresponding to this quantity.
        u_y_ref (ndarray):
        u_y_target (ndarray):
    Returns:

        float: shift in SRF
    """

    # Use spectral database to derive regression yp1 = a + b * y_master, where
    # yp1 corresponds to a shift of Δλ (radiance or BT)

    # then find Δλ that minimises differences

    #bt_ref = srf_master.channel_radiance2bt(srf_master.integrate_radiances(f_spectra, L_spectral_db))

    def fun(Δλ, unit=ureg.um):
        if isinstance(Δλ, numpy.ndarray) and Δλ.ndim > 0:
            if Δλ.size != 1:
                raise ValueError("Δλ must be scalar.  Found "
                    "Δλ.size=={:d}".format(Δλ.size))
            Δλ = Δλ.ravel()[0]
        cost = calc_cost_for_srf_shift(Δλ,
            y_master=y_master, y_target=y_target, srf0=srf0, L_spectral_db=L_spectral_db,
            f_spectra=f_spectra, y_ref=y_ref, unit=unit,
            regression_type=regression_type,
            regression_args=regression_args,
            cost_mode=cost_mode,
            predict_quantity=predict_quantity,
            u_y_ref=u_y_ref,
            u_y_target=u_y_target)
        logging.debug("Shifting {:0<13.7~}: cost {:0<12.8~}".format(
            Δλ*unit, cost))
        return cost.to("1").m

    res = optimiser_func(fun, **optimiser_args)

#    res = scipy.optimize.minimize_scalar(
#            fun=fun,
#            **solver_args)
#            bracket=[-0.1, 0.1],
#            bounds=[-1, 1],
#            method="brent",
#            args=(ureg.um,))
    return res

def filter_outliers(x, cut=5.0):
    """Empirical removal of outliers

    Masks values that are more than *n* times the MAD from the median.

    Arguments:
        
        x (ndarray)
            Array to process

        cut (float)
            Cutoff beyond which to mask.  Defaults to 5.0.
    """

    norm_dist = abs(x - numpy.ma.median(x))/mad(x, masked=True)
    x.mask[norm_dist>=cut] = True

    return x



#!/usr/bin/env python3.3

"""Classes to represent instruments in an abstract way.

Describe a channel by its centre frequency, width, shape, no. of
sidebands, etc., and write out its SRF.  The same for an entire
instrument, which is essentially a collection of channels.

Those are theoretical, idealised instruments, not real instruments with
real SRFs.  For real SRFs, look in physics.SRF.
"""

import datetime
now = datetime.datetime.now

import os.path

import numpy
import decimal
import numbers

from . import physics
from . import tools

class Channel:
    """Represents an idealised (rectangular) radiometer channel.

    """

    def __init__(self, freq, noise, width, sideband=0):
        """Initialise radiometer channel.

        :param freq: Centre frequency [Hz].  This is converted to a
            decimal, and therefore best passed as a string.
        :param noise: Noise level [K]
        :param width: Width of each sideband [Hz].  This is converted to a
            decimal.
        :param sideband: Distance to sidebands (default 0).  This is
            converted to a decimal.
        """
        self.centre_frequency = decimal.Decimal(freq)
        self.noise = noise
        self.width = decimal.Decimal(width)
        self.sideband = decimal.Decimal(sideband)
        self.shape = "rectangular"

    def srf(self, step=99e6):
        """Returns SRF (x, y) with stepsize `step`

        :param step: Stepsize (resolution) [Hz]
        """
        if self.sideband == 0:
            centres = [self.centre_frequency]
        else:
            centres = [self.centre_frequency-self.sideband,
                       self.centre_frequency+self.sideband]
        if self.shape == "rectangular":
            edges = [(float(c-self.width), float(c+self.width)) 
                        for c in centres]
            pp = [numpy.arange(e[0], e[1]+step/2, step) for e in edges]
            f_grid = numpy.concatenate(pp)
            y = numpy.ones(f_grid.size)/f_grid.size
            return (f_grid, y)
        else:
            raise NotImplementedError("Unknown shape: {}".format(self.shape))

    @tools.validator
    def write_srf(self, to, quantity="frequency", factor: numbers.Number=1.0):
        """Write SRF to file.

        :param to: Output file.  Can be an open file, a string containing
            a filename, or a directory.  If it is a directory, the
            filename will be calculated intelligently.

        :param quantity: What unit the SRF is described in.  This should
            be a string, and can be 'frequency' (the defailt)',
            'wavelength', and 'wavenumber'.

        :param float factor: Multiplication factor for the quantity.
            Normally, everything is described in SI units.  However, if
            you want to describe the SRF in cm^-1 instead of m^-1, you
            should pass factor=0.01.
        """
        #tools.validate(Channel.write_srf, locals())
        try:
            to.write
        except AttributeError: # not an open file
            try:
                to = open(to, 'wb')
            except IsADirectoryError:
                to = open(os.path.join(to, "specresp_" +
                                           self.get_chanstr(full=True) +
                                           ".dat"),
                          'wb')

        (x, y) = self.srf()
        if quantity == "frequency":
            pass
        elif quantity == "wavelength":
            x = physics.frequency2wavelength(x)
        elif quantity == "wavenumber":
            x = physics.frequency2wavenumber(x)
        else:
            raise ValueError("Unknown quantity: {}".format(quantity))

        print(now(), "Writing to", to.name)
        numpy.savetxt(to, numpy.c_[x*factor, y])

            
    def get_chanstr(self, full=False):
        """Get the shortest string representation of frequency in GHz

        :param bool full: Be verbose in the channel description.  Defaults
            to False.
        """

        # if whole number, strip off decimal part
        # see also http://stackoverflow.com/q/11227620/974555
        chanstr = str(self.centre_frequency/decimal.Decimal("1e9"))
        if '.' in chanstr:
            chanstr = chanstr.rstrip('0').rstrip('.')

        if full:
            widthstr = "-" + str(self.width/decimal.Decimal("1e9"))
            if self.sideband == 0:
                sbstr = ""
            else:
                sbstr = "±" + str(self.sideband/decimal.Decimal("1e9"))
            return chanstr + sbstr + widthstr
        else:
            return chanstr

    def __repr__(self):
        return "<Channel:{} GHz>".format(self.get_chanstr())

class Radiometer:
    """Represents an ideal radiometer.

    Initialise with list of Channel-objects.
    """

    name = None

    def __init__(self, channels, name="(noname)"):
        """Initialise Radiometer object

        :param channels: List of channels that the radiometer consists of.
        :type channels: List of Channel objects
        :param str name: Instrument name
        """
        self.channels = channels
        self.name = name

    def channel_string(self, pre="", full=False):
        """Get string with channel descriptions.

        Channel descriptions are compact string specifications of channel
        frequencies in GHz. Those are concatenated together separated by
        whitespace.

        :param str pre: String to prepend to each channel.
        :param bool full: Be verbose in the channel descriptions.
            Defaults to False.
        """

        return " ".join(pre+c.get_chanstr(full=full) for c in self.channels)

    def write_all_srf(self, d, *args, **kwargs):
        """Write all SRF to directory d.

        :param str d: Directory to which SRFs will be written.

        Remaining arguments passed on to Channel.write_srf
        """

        for chan in self.channels:
            chan.write_srf(d, *args, **kwargs)

    def __repr__(self):
        if self.name is not None:
            return "<Radiometer {} {} chans>".format(
                        self.name, len(self.channels))
        else:
            return "<Radiometer {} chans>".format(len(self.channels))

# FIXME: read from some external definition file

# icemusic all single band
icemusic = Radiometer([
    Channel("89E9", 0.1, "10E9"), # string to keep precision for Decimal
    Channel("150E9", 0.1, "10E9"),
    Channel("183.31E9", 0.1, "1.8E9"),
    Channel("186.31E9", 0.1, "1.8E9"),
    Channel("190.31E9", 0.1, "1.8E9"),
    Channel("205E9", 0.1, "10E9"),
    Channel("312.5E9", 0.1, "10E9"),
    Channel("321.5E9", 0.1, "3.2E9"),
    Channel("505E9", 0.1, "10E9"),
    Channel("530E9", 0.1, "10E9"),
    Channel("875E9", 0.1, "20E9"),
    Channel("1500E9", 0.1, "20E9")],
        name="icemusic")

# source: buehler07:_ciwsir_qjrms
ciwsir = Radiometer([
    Channel("183.31e9", 0.6, "1.4e9", "1.5e9"),
    Channel("183.31e9", 0.5, "2.0e9", "3.5e9"),
    Channel("183.31e9", 0.4, "3.0e9", "7.0e9"),
    Channel("243.2e9", 0.5, "3.0e9", "2.5e9"),
    Channel("325.15e9", 1.0, "1.6e9", "1.5e9"),
    Channel("325.15e9", 0.8, "2.4e9", "3.5e9"),
    Channel("325.15e9", 0.7, "3.0e9", "9.5e9"),
    Channel("448.0e9", 1.9, "1.2e9", "1.4e9"),
    Channel("448.0e9", 1.4, "2.0e9", "3.0e9"),
    Channel("448.0e9", 1.2, "3.0e9", "7.2e9"),
    Channel("664.0e9", 1.5, "5.0e9", "4.2e9")],
        name="ciwsir")

# source: EPS-SG End User Requirements Document
ici = Radiometer([
    Channel("183.31e9", 0.7, "2.000e9", "7.0e9"),
    Channel("183.31e9", 0.7, "1.500e9", "3.4e9"),
    Channel("183.31e9", 0.7, "1.500e9", "2.0e9"),
    Channel("243.2e9", 0.6, "3.000e9", "2.5e9"),
    Channel("325.15e9", 1.1, "3.000e9", "9.5e9"),
    Channel("325.15e9", 1.2, "2.400e9", "3.5e9"),
    Channel("325.15e9", 1.4, "1.600e9", "1.5e9"),
    Channel("448e9", 1.3, "3.000e9", "7.2e9"),
    Channel("448e9", 1.5, "2.000e9", "3.0e9"),
    Channel("448e9", 1.9, "1.200e9", "1.4e9"),
    Channel("664e9", 1.5, "5.000e9", "4.2e9"),
    ],
        name="ici")

del now



"""Module containing classes abstracting datasets
"""

import abc
import functools
import itertools
import logging
import pathlib
import re
import shelve
import string
import sys
import shutil
import tempfile
import copy

import datetime
import numpy
import numpy.lib.arraysetops
import numpy.lib.recfunctions

import pytz

import typhon.datasets.dataset

from . import tools
from . import time as atmtime
from . import physics
from . import config

class DataFileError(Exception):
    """Superclass for any datafile issues
    """

class InvalidFileError(DataFileError):
    """Raised when the requested information cannot be obtained from the file
    """

class InvalidDataError(DataFileError):
    """Raised when data is not how it should be.
    """

class Dataset(typhon.datasets.dataset.Dataset):
    timezone = pytz.UTC

    def __init__(self, *, memory=None, **kwargs):
        # set memorisation on methods.  See note on "caching methods" at
        # https://pythonhosted.org/joblib/memory.html
        tools.setmem(self, memory)
        super().__init__(**kwargs)

    @tools.mark_for_disk_cache(
        process=dict(
            my_data=lambda x: x.view(dtype="i1")))
    def combine(self, my_data, other_obj):
        return super().combine(my_data, other_obj)

    @staticmethod
    def extend_with_doy_localtime(M):
        """Calculate DOY and mean local solar time

        :param M: ndarray with dtype as returned by self.read 
        :returns M: ndarray with dtype extended with doy and mlst
        """

        (all_doy, all_mlst) = (numpy.array(x) for x in zip(
            *[atmtime.dt_to_doy_mlst(M[i]["time"].astype(datetime.datetime),
                                     M[i]["lon"])
                for i in range(M.shape[0])]))
        return numpy.lib.recfunctions.append_fields(M, 
            names=["doy", "mlst"],
            data=[all_doy, all_mlst],
            dtypes=["u2", "f4"])
    
    @staticmethod
    def extend_with_dofs(M):
        """Calculate DOFs

        :param M: ndarray with dtype as for self.read
        :returns M: M with added field dof
        """

        return numpy.lib.recfunctions.append_fields(M,
            names=["dof"],
            data=[physics.AKStats(M["ch4_ak"],
                    name="DUMMY").dofs()],
            dtypes=["f4"])

    def flag(self, arr):
        """Must be implemented by child
        """
        return arr

class SingleFileDataset(Dataset, typhon.datasets.dataset.SingleFileDataset):
    pass

class MultiFileDataset(Dataset, typhon.datasets.dataset.MultiFileDataset):
    pass

class SingleMeasurementPerFileDataset(MultiFileDataset, typhon.datasets.dataset.SingleMeasurementPerFileDataset):

    _head_dtype = dict(CH4_total="<f4")
    def _read(self, p, fields="all"):
        super()._read(p, fields)

class HomemadeDataset(MultiFileDataset, typhon.datasets.dataset.HomemadeDataset):
    pass

class ProfileDataset(Dataset):
    """Abstract superclass with stuff for anything having profiles

    Additional attributes compared to its parent::

    - range::

        If applicable, a tuple with (lo, hi) vertical limits of
        sensitivity range in metre.

    """

    # source of profile sizes.  Can be a number (fixed size) or a string
    # (field name to get size from)
    n_prof = "p"

    # does the A priori need converting?  It does for PEARL, does not for
    # others
    A_needs_converting = tools.NotTrueNorFalse
    A_needs_swapping = tools.NotTrueNorFalse

    range = None

    #@tools.mutable_cache(maxsize=10)
    def read(self, f=None, fields="all"):
        M = super().read(f, fields)
        if isinstance(self.n_prof, str):
            n_prof = M.dtype[self.n_prof].shape
        else:
            n_prof = self.n_prof
        if not "z" in M.dtype.names:
            
            newM = numpy.empty(shape=M.shape,
                dtype=M.dtype.descr + [("z", "f4", n_prof)])
            for nm in M.dtype.names:
                newM[nm] = M[nm]
            logging.debug("Adding {:d} z-profiles".format(M.shape[0]))
            for i in range(M.shape[0]):
                newM["z"][i, :] = self.get_z(M[i])
            M = newM
        return M

    def get_z(self, dt):
        """Get z-profile in metre.

        Takes as argument a single measurement, with the dtype returned by
        its own reading routine.
        """
        return dt["z"]

    def get_z_for(self, dt, field):
        """Get z-profile for particular field.

        Takes as argument a single measurement and a field name.  This is
        different from get_z because sometimes, some fields are on
        different grids (such as T for Tanso).
        """
        return self.get_z(dt)

class StationaryDataset(Dataset):
    """Abstract superclass for any ground-station dataset
    """

    unique_fields = {"time"}



import subprocess
import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def get_version():
    """Obtain version number
    """
    #from pyatmlab.meta import get_version as get_v
    from pyatmlab import __version__ as v
    if v.endswith("+"):
        #cp = subprocess.run(["git", "log", "--format=%H", "-n", "1"], stdout=subprocess.PIPE, check=True)
        cp = subprocess.run(["git", "rev-parse", "--short", "HEAD"],
                            stdout=subprocess.PIPE, check=True)
        v += cp.stdout.decode("ascii").strip()
    return v
    #return get_v(os.path.join(os.path.dirname(__file__), "ChangeLog"))

setup(
    name = "pyatmlab",
    version = get_version(),
    author = "Gerrit Holl",
    author_email = "g.holl@reading.ac.uk",
    description = ("Diverse collection of tools for working with "
                   "atmospheric data from remote sensing, modelling, "
                   "and related applications."),
    license = "Modified BSD License",
    keywords = "atmosphere, satellite, remote sensing, radiative transfer",
    url = "http://packages.python.org/pyatmlab",
    packages=['pyatmlab', 'tests', 'pyatmlab.datasets', 'pyatmlab.arts'],
    install_requires=["numpy>=1.10", "scipy>=0.16", "pyproj>=1.9",
                      "statsmodels>=0.6", "pytz>=2015.4",
                      "progressbar2>=3.3",
                      "matplotlib>=1.4", "pint>=0.6",
                      "numexpr>=2.4",
                      "h5py>=2.5",
                      "netCDF4>=1.1",
                      "scikit-learn>=0.17",
                      "typhon>=0.2.7"],
    long_description=read('README'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
)
